package cmd

import (
	"runtime"

	"github.com/rs/zerolog"

	"orus.io/orus-io/rednerd/engines/libreoffice"
	"orus.io/orus-io/rednerd/engines/mjml"
	"orus.io/orus-io/rednerd/engines/mustache"
	"orus.io/orus-io/rednerd/engines/od"
	"orus.io/orus-io/rednerd/engines/wpd"
	"orus.io/orus-io/rednerd/rendering"
)

var engineOptions EngineOptions

type EngineOptions struct {
	LibreOffice libreoffice.EngineOptions `group:"LibreOffice" namespace:"libreoffice" env-namespace:"LIBREOFFICE" ini-name:"libreoffice"`
	Wpd         wpd.EngineOptions         `group:"WPD" namespace:"wpd" env-namespace:"WPD" ini-name:"wpd"`
}

// SetupEngines creates and register the rendering engines.
func (o *EngineOptions) SetupEngines(log zerolog.Logger) error {
	if o.Wpd.WorkerPoolSize == 0 {
		o.Wpd.WorkerPoolSize = runtime.NumCPU()
	}
	if err := rendering.RegisterPooledEngine(
		"wpd", func(log zerolog.Logger) (rendering.Engine, error) {
			return wpd.NewEngine(o.Wpd, log)
		}, o.Wpd.WorkerPoolSize, log); err != nil {
		return err
	}

	if err := rendering.RegisterTemplateEngine("mustache", &mustache.Engine{}); err != nil {
		return err
	}

	mjmlEngine, err := mjml.NewEngine()
	if err != nil {
		return err
	}
	if err := rendering.RegisterEngine("mjml", mjmlEngine); err != nil {
		return err
	}

	odEngine := od.NewEngine()
	if err := rendering.RegisterTemplateEngine("od", odEngine); err != nil {
		return err
	}

	loEngine, err := libreoffice.NewEngine(o.LibreOffice, log)
	if err != nil {
		return err
	}

	if err := rendering.RegisterEngine("lokit", loEngine); err != nil {
		return err
	}

	return nil
}
