package wpd

import (
	"bufio"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"sync"
	"sync/atomic"
	"time"

	"github.com/rs/zerolog"
)

var (
	ErrJobFailed = errors.New("wpd render job failed")
	ErrWpdFailed = errors.New("wpd crashed")
)

// Request ...
type Request struct {
	ID   uint64 `json:"id"`
	CSS  string `json:"css"`
	Size int    `json:"size"`
	HTML []byte `json:"-"`
}

// Response ...
type Response struct {
	ID      uint64 `json:"id"`
	Content []byte `json:"-"`
	Size    int    `json:"size"`
	Error   string `json:"error"`
}

type Info struct {
	Wpd      string `json:"wpd"`
	Version  string `json:"version"`
	Protocol string `json:"protocol"`
}

type WPD interface {
	Submit(ctx context.Context, css string, html []byte) (uint64, error)
	Cancel(uint64) error
	// Wait for a result to be returned by wpd and returns it.
	// if wpd gets stopped, must return a io.EOF
	NextResult() (uint64, []byte, error)
}

type processWPD struct {
	process Process
	reader  *bufio.Reader
	counter uint64
	m       sync.Mutex
}

func NewProcessWPD(p Process, log zerolog.Logger) (WPD, error) {
	reader := bufio.NewReader(p.Stdout())
	info, err := helo(p, reader, log)
	if err != nil {
		if err := p.Stop(); err != nil {
			log.Err(err).Msg("problem stopping the runner")
		}

		return nil, fmt.Errorf("wpd helo failed: %w", err)
	}
	if info.Protocol != "1" {
		if err := p.Stop(); err != nil {
			log.Err(err).Msg("problem stopping the runner")
		}

		return nil, fmt.Errorf("invalid wpd protocol version. Expects \"1\", got: \"%s\"", info.Protocol)
	}

	return &processWPD{
		process: p,
		reader:  reader,
	}, nil
}

func helo(p Process, reader *bufio.Reader, log zerolog.Logger) (info Info, err error) {
	t := time.AfterFunc(time.Second*5, func() {
		if err := p.Stop(); err != nil {
			log.Err(err).Msg("Error stopping the runner")
		}
	})
	defer func() {
		if !t.Stop() {
			err = fmt.Errorf("timeout reading runner info header")
		}
	}()

	line, err := reader.ReadBytes(byte('\n'))
	if err != nil {
		return info, fmt.Errorf("could not read from wpd process: %w", err)
	}
	if err := json.Unmarshal(line, &info); err != nil {
		panic(err)
	}
	if info.Wpd == "" {
		return info, fmt.Errorf("invalid wpd info: missing \"wpd\"")
	}

	return info, nil
}

func (p *processWPD) sendRequest(request Request) error {
	p.m.Lock()
	defer p.m.Unlock()

	if err := json.NewEncoder(p.process.Stdin()).Encode(request); err != nil {
		return err
	}
	if _, err := WriteAll(
		p.process.Stdin(),
		request.HTML, []byte("\n"),
	); err != nil {
		return err
	}

	return nil
}

func (p *processWPD) sendCancel(id uint64) error {
	p.m.Lock()
	defer p.m.Unlock()

	if err := json.NewEncoder(p.process.Stdin()).Encode(map[string]interface{}{
		"id":     id,
		"action": "cancel",
	}); err != nil {
		return err
	}

	return nil
}

func (p *processWPD) Submit(ctx context.Context, css string, html []byte) (uint64, error) {
	id := atomic.AddUint64(&p.counter, 1)
	err := p.sendRequest(Request{
		ID:   id,
		CSS:  css,
		Size: len(html),
		HTML: html,
	})
	if err != nil {
		return 0, err
	}

	return id, nil
}

func (p *processWPD) Cancel(id uint64) error {
	return p.sendCancel(id)
}

func (p *processWPD) receiveResponse() (*Response, error) {
	line, err := p.reader.ReadBytes(byte('\n'))
	if err != nil {
		return nil, fmt.Errorf("could not read response: %w", err)
	}
	var response Response
	if err := json.Unmarshal(line, &response); err != nil {
		return nil, fmt.Errorf("could not unmarshal response: %w", err)
	}
	if response.Size != 0 {
		buf := make([]byte, response.Size)
		_, err := io.ReadFull(p.reader, buf)
		if err != nil {
			return nil, fmt.Errorf("could not read response body: %w", err)
		}

		if b, err := p.reader.ReadByte(); err != nil || b != byte('\n') {
			if err != nil {
				return nil, fmt.Errorf("could not read response body final CR: %w", err)
			}

			return nil, fmt.Errorf("response body final CR is: '%b'", b)
		}
		response.Content = buf
	}

	return &response, nil
}

func (p *processWPD) NextResult() (uint64, []byte, error) {
	response, err := p.receiveResponse()
	if err != nil {
		return response.ID, nil, fmt.Errorf("%w: %s", ErrWpdFailed, err)
	}
	if response.Size == 0 {
		if response.Error == "" {
			response.Error = "<missing error>"
		}

		return response.ID, nil, fmt.Errorf("%w: wpd job %d error: %s", ErrJobFailed, response.ID, response.Error)
	}

	return response.ID, response.Content, nil
}

// WriteAll writes a list of []byte to a writer in one shot.
func WriteAll(w io.Writer, data ...[]byte) (int, error) {
	var n int
	for _, d := range data {
		nn, err := w.Write(d)
		n += nn
		if err != nil {
			return n, err
		}
	}

	return n, nil
}
