package handlers

import (
	"context"
	"net/http"

	"github.com/go-openapi/runtime"
	"github.com/go-openapi/runtime/middleware"

	"orus.io/orus-io/rednerd/auth"
	"orus.io/orus-io/rednerd/models"
	op_auth "orus.io/orus-io/rednerd/restapi/operations/auth"
)

// NewAuthAuthenticateHandler returns a Authenticate instance.
func NewAuthAuthenticateHandler(auth *auth.Auth) *AuthAuthenticate {
	return &AuthAuthenticate{auth}
}

// AuthAuthenticate handles /auth/authenticate.
type AuthAuthenticate struct {
	auth *auth.Auth
}

func (h *AuthAuthenticate) handle(
	ctx context.Context, credentials *models.Credentials,
) (string, *http.Cookie, error) {
	principal, err := h.auth.Authenticate(ctx, credentials)
	if err != nil {
		return "", nil, err
	}
	token, cookie, err := h.auth.MakeToken(*principal, credentials.SetCookie)
	if err != nil {
		return "", nil, err
	}

	return token, cookie, nil
}

// Handle authenticates the user with the given credential and returns a JWT.
func (h *AuthAuthenticate) Handle(
	params op_auth.AuthenticateParams,
) middleware.Responder {
	token, cookie, err := h.handle(params.HTTPRequest.Context(), params.Credentials)
	if err != nil {
		switch err {
		case auth.ErrInvalidCredentials:
			return op_auth.
				NewAuthenticateUnauthorized().
				WithPayload(models.FromError(err))
		default:
			return op_auth.
				NewAuthenticateDefault(http.StatusInternalServerError).
				WithPayload(models.FromError(err))
		}
	}

	return NewAuthenticateOK().WithPayload(token).WithCookie(cookie)
}

// Below is a customized AuthenticateOK copied from the swagger-generated files
// The only difference is that it handles writing a cookie to the response

// AuthenticateOKCode is the HTTP code returned for type AuthenticateOK.
const AuthenticateOKCode int = 200

/*
AuthenticateOK A Json Web Token with a 5 minutes validity

swagger:response authenticateOK
*/
type AuthenticateOK struct {
	/*
	  In: Body
	*/
	Payload string `json:"body,omitempty"`

	Cookie *http.Cookie
}

// NewAuthenticateOK creates AuthenticateOK with default headers values.
func NewAuthenticateOK() *AuthenticateOK {
	return &AuthenticateOK{}
}

// WithPayload adds the payload to the authenticate o k response.
func (o *AuthenticateOK) WithPayload(payload string) *AuthenticateOK {
	o.Payload = payload

	return o
}

// SetPayload sets the payload to the authenticate o k response.
func (o *AuthenticateOK) SetPayload(payload string) {
	o.Payload = payload
}

// WithCookie adds the cookie to the authenticate OK response.
func (o *AuthenticateOK) WithCookie(cookie *http.Cookie) *AuthenticateOK {
	o.Cookie = cookie

	return o
}

// WriteResponse to the client.
func (o *AuthenticateOK) WriteResponse(rw http.ResponseWriter, producer runtime.Producer) {
	if o.Cookie != nil {
		http.SetCookie(rw, o.Cookie)
	}
	rw.WriteHeader(200)
	payload := o.Payload
	if err := producer.Produce(rw, payload); err != nil {
		panic(err) // let the recovery middleware deal with this
	}
}
