package handlers

import (
	"encoding/csv"
	"fmt"
	"io"
	"net/http"
	"path"
	"strconv"
	"strings"
	"time"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"golang.org/x/net/context"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/admin"
)

// NewRenderLogsHandler creates a RenderLogs handler.
func NewRenderLogsHandler(db *sqlx.DB) *RenderLogs {
	return &RenderLogs{db}
}

type Period struct {
	From time.Time
	To   time.Time
}

func parsePeriod(s string) (*Period, error) {
	if len(s) != 7 || s[4] != '-' {
		return nil, fmt.Errorf("invalid period. Expects YYYY-MM")
	}
	year, err := strconv.Atoi(s[:4])
	if err != nil {
		return nil, fmt.Errorf("invalid period. Error parsing the year: %w", err)
	}
	month, err := strconv.Atoi(s[5:])
	if err != nil {
		return nil, fmt.Errorf("invalid period. Error parsing the month: %w", err)
	}
	toYear := year
	toMonth := month + 1
	if toMonth > 12 {
		toMonth -= 12
		toYear++
	}

	return &Period{
		time.Date(year, time.Month(month), 1, 0, 0, 0, 0, time.Local),
		time.Date(toYear, time.Month(toMonth), 1, 0, 0, 0, 0, time.Local),
	}, nil
}

// RenderLogs is a handler for /renderlogs.
type RenderLogs struct {
	db *sqlx.DB
}

func (h *RenderLogs) handle(ctx context.Context, period Period) (io.ReadCloser, error) {
	log := zerolog.Ctx(ctx)
	q := squirrel.Select(models.RenderLogColumns...).
		From(models.RenderLogTable).
		Where(squirrel.And{
			squirrel.GtOrEq{
				models.RenderLogRenderedAtColumn: period.From,
			}, squirrel.Lt{
				models.RenderLogRenderedAtColumn: period.To,
			},
		})

	var logs []models.RenderLog

	pipeOut, pipeIn := io.Pipe()

	go func() {
		defer pipeIn.Close()

		db := database.NewSQLHelper(ctx, h.db, *log)

		if err := db.Select(&logs, q); err != nil {
			pipeIn.CloseWithError(err)
		}

		writer := csv.NewWriter(pipeIn)
		writer.UseCRLF = true

		if err := writer.Write([]string{
			"rendered_at", "duration", "from_type", "to_type", "username", "template_language",
		}); err != nil {
			log.Err(err).Msg("RenderLogs: failed to write csv header")

			return
		}

		for _, rlog := range logs {
			if err := writer.Write([]string{
				rlog.RenderedAt.Format(time.RFC3339),
				fmt.Sprintf("%d", rlog.Duration),
				rlog.FromType,
				rlog.ToType,
				rlog.Username,
				rlog.TemplateLanguage.String,
			}); err != nil {
				log.Err(err).Msg("RenderLogs: failed to write csv line")

				return
			}
		}

		writer.Flush()
		if err := writer.Error(); err != nil {
			log.Err(err).Msg("RenderLogs: failed to flush csv file")
		}
	}()

	return pipeOut, nil
}

func (h *RenderLogs) Handle(
	params admin.RenderlogsParams, principal *models.Principal,
) middleware.Responder {
	log := zerolog.Ctx(params.HTTPRequest.Context())
	if !principal.IsAdmin {
		return admin.NewRenderlogsUnauthorized()
	}

	ext := path.Ext(params.Period)
	params.Period = strings.TrimSuffix(params.Period, ext)

	if ext != ".csv" {
		return admin.NewRenderlogsNotFound().
			WithPayload(&models.Error{Message: "extension must be '.csv'"})
	}

	period, err := parsePeriod(params.Period)
	if err != nil {
		return admin.NewRenderlogsNotFound().
			WithPayload(&models.Error{Message: "period format must be 'AAAA-MM'"})
	}

	reader, err := h.handle(params.HTTPRequest.Context(), *period)
	if err != nil {
		return InternalError(err, log, false)
	}

	return NewRenderlogsOK().
		WithContentType("text/csv").
		WithContentDisposition("attachment; filename=" + params.Period + ".csv").
		WithPayload(reader)
}

type RenderlogsOK struct {
	/*

	 */
	ContentDisposition string `json:"Content-Disposition"`
	/*

	 */
	ContentType string `json:"Content-Type"`

	/*
	  In: Body
	*/
	Payload io.ReadCloser `json:"body,omitempty"`
}

// Home made responder

// NewRenderlogsOK creates RenderlogsOK with default headers values.
func NewRenderlogsOK() *RenderlogsOK {
	return &RenderlogsOK{}
}

// WithContentDisposition adds the contentDisposition to the renderlogs o k response.
func (o *RenderlogsOK) WithContentDisposition(contentDisposition string) *RenderlogsOK {
	o.ContentDisposition = contentDisposition

	return o
}

// SetContentDisposition sets the contentDisposition to the renderlogs o k response.
func (o *RenderlogsOK) SetContentDisposition(contentDisposition string) {
	o.ContentDisposition = contentDisposition
}

// WithContentType adds the contentType to the renderlogs o k response.
func (o *RenderlogsOK) WithContentType(contentType string) *RenderlogsOK {
	o.ContentType = contentType

	return o
}

// SetContentType sets the contentType to the renderlogs o k response.
func (o *RenderlogsOK) SetContentType(contentType string) {
	o.ContentType = contentType
}

// WithPayload adds the payload to the renderlogs o k response.
func (o *RenderlogsOK) WithPayload(payload io.ReadCloser) *RenderlogsOK {
	o.Payload = payload

	return o
}

// SetPayload sets the payload to the renderlogs o k response.
func (o *RenderlogsOK) SetPayload(payload io.ReadCloser) {
	o.Payload = payload
}

// WriteResponse to the client.
func (o *RenderlogsOK) WriteResponse(rw http.ResponseWriter, producer runtime.Producer) {
	// response header Content-Disposition

	contentDisposition := o.ContentDisposition
	if contentDisposition != "" {
		rw.Header().Set("Content-Disposition", contentDisposition)
	}

	// response header Content-Type

	contentType := o.ContentType
	if contentType != "" {
		rw.Header().Set("Content-Type", contentType)
	}

	rw.WriteHeader(200)
	payload := o.Payload
	if _, err := io.Copy(rw, payload); err != nil {
		panic(err)
	}

	_ = payload.Close()
}
