package tests

import (
	"bytes"
	"encoding/base64"
	"encoding/json"
	"io/ioutil"
	"net/http"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"orus.io/orus-io/od-templating"

	"orus.io/orus-io/rednerd/models"
)

func TestRender(t *testing.T) {
	tester := NewAPITester(t, nil)
	defer tester.Close()

	defer tester.Logout()
	tester.Login("admin", "admin")

	tester.CreateUser("janedoe", "janedoe")
	tester.Login("janedoe", "janedoe")

	t.Run("MJML-Template-to-html", func(t *testing.T) {
		defer tester.SetT(t)()
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"template": JSONObj{
					"language":    "mustache",
					"produces":    "text/mjml",
					"body-format": "text",
					"body":        "<mjml> <mj-body> <mj-section> <mj-column> <mj-text>hello {{name}}</mj-text> </mj-column> </mj-section> </mj-body> </mjml>",
				},
				"data": []JSONObj{{"name": "Nath"}},
			},
			&result)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello Nath")
		}
	})

	t.Run("MJML-Saved-Template-to-html", func(t *testing.T) {
		defer tester.SetT(t)()
		require.Equal(t, http.StatusCreated,
			tester.MustPostJSON(
				"/api/v1/template/janedoe",
				JSONObj{
					"name":        "t1",
					"language":    "mustache",
					"produces":    "text/mjml",
					"body-format": "text",
					"body":        "<mjml> <mj-body> <mj-section> <mj-column> <mj-text>hello {{name}}</mj-text> </mj-column> </mj-section> </mj-body> </mjml>",
				},
				nil),
		)
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"template": JSONObj{
					"account": "janedoe",
					"name":    "t1",
				},
				"data": []JSONObj{{"name": "Nath"}},
			},
			&result)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello Nath")
		}
	})

	t.Run("MJML-with-errors", func(t *testing.T) {
		defer tester.SetT(t)()
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"document": JSONObj{
					"type":        "text/mjml",
					"body-format": "test",
					"body": `<mjml>
							  <mj-body>
								<mj-section invalid="hi">
								  <mj-column>
									<mj-text invalid=true>hello world</mj-text>
								  </mj-column>
								</mj-section>
							  </mj-body>
							</mjml>`,
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello world")
		}
		t.Log(result[0].RenderErrors)
		require.Equal(t, 6, len(result[0].RenderErrors))
		assert.Equal(t, "MJML", result[0].RenderErrors[0].Engine)
		assert.Equal(t, "MJML", result[0].RenderErrors[1].Engine)
	})
	t.Run("MJML-skip-validation", func(t *testing.T) {
		defer tester.SetT(t)()
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"metadata": JSONObj{
					"mjml-validation-level": "skip",
				},
				"document": JSONObj{
					"type":        "text/mjml",
					"body-format": "test",
					"body": `<mjml>
							  <mj-body>
								<mj-section invalid="hi">
								  <mj-column>
									<mj-text invalid=true>hello world</mj-text>
								  </mj-column>
								</mj-section>
							  </mj-body>
							</mjml>`,
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello world")
		}
		require.Equal(t, 0, len(result[0].RenderErrors))
	})

	t.Run("OD-render", func(t *testing.T) {
		defer tester.SetT(t)()
		b, err := ioutil.ReadFile("../dependencies/od-templating/examples/basics.odt")
		require.NoError(t, err)
		d, err := ioutil.ReadFile("../dependencies/od-templating/examples/basics.json")
		require.NoError(t, err)
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "application/vnd.oasis.opendocument.text",
				"template": JSONObj{
					"language":    "od+mustache",
					"produces":    "application/vnd.oasis.opendocument.text",
					"body-format": "base64",
					"body":        base64.StdEncoding.EncodeToString(b),
				},
				"data": []json.RawMessage{json.RawMessage(d)},
			},
			&result,
		)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			expect, err := od.LoadDocumentFilename("basics-expects.odt")
			require.NoError(t, err)
			bodyDoc, err := od.LoadDocument(bytes.NewReader(body), int64(len(body)))
			require.NoError(t, err)
			expectContent, err := expect.Content()
			require.NoError(t, err)
			bodyContent, err := bodyDoc.Content()
			require.NoError(t, err)
			assert.Equal(t, string(expectContent), string(bodyContent))
		}
		require.Equal(t, 0, len(result[0].RenderErrors))
	})
	t.Run("libreoffice", func(t *testing.T) {
		defer tester.SetT(t)()
		b, err := ioutil.ReadFile("../dependencies/od-templating/examples/basics.odt")
		require.NoError(t, err)
		d, err := ioutil.ReadFile("../dependencies/od-templating/examples/basics.json")
		require.NoError(t, err)
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "application/pdf",
				"template": JSONObj{
					"language":    "od+mustache",
					"produces":    "application/vnd.oasis.opendocument.text",
					"body-format": "base64",
					"body":        base64.StdEncoding.EncodeToString(b),
				},
				"data": []json.RawMessage{json.RawMessage(d)},
			},
			&result,
		)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			require.NoError(t, ioutil.WriteFile("basics.pdf", body, 0o600))
		}
		require.Equal(t, 0, len(result[0].RenderErrors))
	})
	t.Run("MJML-specific-test", func(t *testing.T) {
		defer tester.SetT(t)()
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"document": JSONObj{
					"type":        "text/mjml",
					"body-format": "text",
					"body":        mjmlSource,
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "<strong>Client</strong> : client")
		}
	})
}

const mjmlSource = `
<mjml>

  <mj-head>

    <mj-attributes>

	<mj-all font-family="Arial" font-size="16px" />
	<mj-text color="#000" line-height="28px" />

	<mj-class name="main-title" font-size="22px" font-weight="bold" color="#969494" />
	<mj-class name="subtitle" font-weight="bold" />
	<mj-class name="primary-color" color="#C32427" />
	<mj-class name="secondary-color" color="#003E8B" />
	<mj-class name="small-italic" font-style="italic" font-size="14px" line-height="18px"/>

	<mj-class name="divider" border-width="1px" border-style="solid" border-color="lightgrey" />
	<mj-class name="border-content" border="1px solid #dedede" padding-top="10px" padding-bottom="10px" />
	<mj-class name="cta-btn" inner-padding="22px 44px" background-color="#003E8B" color="white" font-size="14px" />

	<mj-class name="borded-left" border-left="1px solid #dedede" />
	<mj-class name="borded-left__text" padding="5px 20px" />

	<mj-class name="no-vertical-padding" padding-top="0" padding-bottom="0" />

</mj-attributes>

<mj-style inline="inline">

	a{
		color: #000;
	}
	.link-white {
		color:#fff;
		text-decoration: none;
	}
	.mj-table th, .mj-table td{
		border: 1px solid #ecedee;
		text-align: left;
		padding: 8px 16px;
	}
	.secondary-color{
		color: #003E8B !important;
	}
	.bold{
		font-weight: bold !important;
	}

</mj-style>

<mj-style>

	@media all and (max-width: 480px) { .mobile-center *{ text-align: center !important; }

</mj-style>

  </mj-head>

  <mj-body background-color="#f5f5f5" width="650px">

    <mj-section>

  <mj-column width="30%">
  </mj-column>

  <mj-column width="70%">
    <mj-text font-size="16px" font-weight="bold" align="right" css-class="mobile-center">
      JAL-PRO - ORDRE DE PUBLICATION
    </mj-text>
  </mj-column>

</mj-section>

    <mj-section background-color="#fff">

      <mj-column>

        <mj-wrapper padding="0 20px" background-color="#fff">
          <mj-section>
            <mj-column mj-class="border-content">
              <mj-text align="left">
                <strong>Client</strong> : client<br>
              </mj-text>
            </mj-column>
          </mj-section>
        </mj-wrapper>

      </mj-column>
    </mj-section>
</mj-wrapper>

  </mj-body>

</mjml>

`
